import inspect
from asyncio import iscoroutinefunction
from collections.abc import Awaitable, Callable, Coroutine
from typing import Any, Protocol, TypeAlias, TypeGuard

from fastapi import Depends
from htmy import as_component_type

from holm.fastapi import FastAPIDependency


class SyncLayout(Protocol):
    """
    Sync layout protocol definition.
    """

    def __call__(self, __props: Any, /, **dependencies: Any) -> Any: ...


class AsyncLayout(Protocol):
    """
    Async layout protocol definition.
    """

    async def __call__(self, __props: Any, /, **dependencies: Any) -> Any: ...


Layout: TypeAlias = SyncLayout | AsyncLayout
"""
Layout type definition.

A layout is a callable that expects a single positional argument the component or data returned
by the wrapped layout or page, and returns the properties for its wrapper layout. The root layout
must always return a `Component`.
"""

LayoutFactory: TypeAlias = Callable[[Any], Any | Coroutine[None, None, Any]]
"""
A layout factory is a sync or asynccallable that expects a single argument (the layout's properties)
and returns the properties for its wrapper layout.

The root layout factory must always return a `htmy` `Component`.
"""


class LayoutDefinition(Protocol):
    """Protocol definition for objects (usually modules) that define a layout."""

    @property
    def layout(self) -> Layout: ...


def is_layout_definition(obj: Any) -> TypeGuard[LayoutDefinition]:
    """Type guard for `LayoutDefinition`."""
    layout = getattr(obj, "layout", None)
    return callable(layout)


def combine_layouts_to_dependency(
    outer_dep: FastAPIDependency[LayoutFactory], inner: Layout | None
) -> FastAPIDependency[LayoutFactory]:
    """
    Creates a layout factory dependency that wraps the inner layout inside the outer layout.

    Arguments:
        outer: The outer layout factory.
        inner: The inner layout factory.
    """
    if inner is None:
        return outer_dep

    inner_dep = layout_to_dependency(inner)

    async def combined_layout_dep(
        outer: LayoutFactory = Depends(outer_dep),  # noqa: B008
        inner: LayoutFactory = Depends(inner_dep),  # noqa: B008
    ) -> LayoutFactory:
        async def layout_factory(props: Any) -> Any:
            inner_result = inner(props)
            if isinstance(inner_result, Awaitable):
                inner_result = await inner_result

            result = outer(as_component_type(inner_result))
            if isinstance(result, Awaitable):
                result = await result

            return result

        return layout_factory

    return combined_layout_dep


def empty_layout(props: Any) -> Any:
    """`LayoutFactory` that simply returns the received props."""
    return props


def empty_layout_dependency() -> FastAPIDependency[LayoutFactory]:
    """
    FastAPI dependency that returns `empty_layout`.
    """
    return empty_layout


def layout_to_dependency(layout: Layout) -> FastAPIDependency[LayoutFactory]:
    """
    Converts a layout to a FastAPI dependency that returns a layout factory.
    """
    # eval_str=True is necessary in case future __annotations__ is used
    # where the layout is function is defined.
    params = tuple(inspect.signature(layout, eval_str=True).parameters.values())
    props_index = next((i for i, p in enumerate(params) if p.name != "self"), -1)
    if props_index == -1:
        raise ValueError("Layout factory must at least one argument.")

    result: FastAPIDependency[LayoutFactory]
    if iscoroutinefunction(layout):

        async def async_layout_factory_dep(**kwargs: Any) -> LayoutFactory:
            async def layout_factory(props: Any) -> Any:
                return await layout(props, **kwargs)

            return layout_factory

        result = async_layout_factory_dep
    else:

        def sync_layout_factory_dep(**kwargs: Any) -> LayoutFactory:
            def layout_factory(props: Any) -> Any:
                return layout(props, **kwargs)

            return layout_factory

        result = sync_layout_factory_dep

    # Copy the dependencies part of the signature from the original function, so
    # FastAPI can resolve those arguments. The props argument must be excluded.
    result.__signature__ = inspect.Signature(params[props_index + 1 :])  # type: ignore[union-attr]

    return result



__version__ = "0.4.0"

from .app import App as App
from .fastapi import FastAPIDependency as FastAPIDependency
from .fastapi import FastAPIErrorHandler as FastAPIErrorHandler
from .module_options._metadata import Metadata as Metadata
from .modules._error import ErrorHandlerMapping as ErrorHandlerMapping



import inspect
from collections.abc import Awaitable, Iterable
from functools import lru_cache
from pathlib import Path
from typing import Any, cast

from fastapi import APIRouter, Depends, FastAPI, Response
from fasthx.htmy import HTMY
from htmy import Component, as_component_type

from ._model import AppConfig, AppNode, PackageInfo
from .fastapi import FastAPIDependency
from .module_options._metadata import (
    MetadataMapping,
    components_with_metadata,
    empty_metadata_dep,
    get_metadata_dependency,
)
from .module_options._submit_handler import get_submit_handler
from .modules._api import PlainAPIFactory, RenderingAPIFactory, is_api_definition
from .modules._error import is_error_handler_owner, register_error_handlers
from .modules._layout import (
    LayoutFactory,
    combine_layouts_to_dependency,
    empty_layout_dependency,
    is_layout_definition,
)
from .modules._page import is_page_definition
from .typing import module_names


def App(*, app: FastAPI | None = None, htmy: HTMY | None = None) -> FastAPI:
    """
    Creates a FastAPI application with all the routes that are defined in the application package.

    Arguments:
        app: Optional FastAPI application to use. If `None`, a new instance will be created.
        htmy: Optional `fasthx.htmy.HTMY` instance to use for server-side rendering. If `None`,
            a default instance will be created.
    """
    if app is None:
        app = FastAPI()

    if htmy is None:
        htmy = HTMY()

    config = AppConfig.default()
    packages = _discover_app_packages(config)
    root_node = _build_app_tree(packages)

    # Register error handlers
    if pkg := root_node.package:
        register_error_handlers(app, pkg.import_module("error", is_error_handler_owner), htmy=htmy)

    # Build the API
    app.include_router(_build_api(root_node, htmy=htmy))

    return app


def _build_api(
    node: AppNode,
    *,
    base_layout_dep: FastAPIDependency[LayoutFactory] = empty_layout_dependency,
    htmy: HTMY,
) -> APIRouter:
    """
    Recursively builds an `APIRouter` based on the application defined by `node`.

    Arguments:
        node: Application definition.
        base_layout: The base layout dependency to use for the API.
        htmy: The `fasthx.htmy.HTMY` instance to use for server-side rendering.
    """
    layout_dep = base_layout_dep  # In case there is no package or it has no layout module.
    pkg = node.package
    api = _make_api_router_for_package(pkg, htmy)
    if pkg is not None:
        layout_module = pkg.import_module("layout", is_layout_definition)
        page_module = pkg.import_module("page", is_page_definition)

        layout_dep = combine_layouts_to_dependency(
            base_layout_dep, None if layout_module is None else layout_module.layout
        )
        page_dep, submit_handler_dep, metadata_dep = (
            (None, None, None)
            if page_module is None
            else (
                page_module.page,
                get_submit_handler(page_module),
                get_metadata_dependency(page_module),
            )
        )

        if page_dep is not None:
            path_operation = _make_page_path_operation(
                layout_dep=layout_dep,
                metadata_dep=empty_metadata_dep if metadata_dep is None else metadata_dep,
                page_dep=page_dep,
            )

            # Register the route with rendering.
            api.get(
                "/",
                response_model=None,
                name="page",
                description=page_dep.__doc__,
                tags=["Page"],
            )(htmy.page(components_with_metadata)(path_operation))

        if submit_handler_dep is not None:
            path_operation = _make_page_path_operation(
                layout_dep=layout_dep,
                metadata_dep=empty_metadata_dep if metadata_dep is None else metadata_dep,
                page_dep=submit_handler_dep,
            )

            # Register the route with rendering.
            api.post(
                "/",
                response_model=None,
                name="submit",
                description=submit_handler_dep.__doc__,
                tags=["Page", "Submit"],
            )(htmy.page(components_with_metadata)(path_operation))

    for sub_url, child_node in node.subtree.items():
        api.include_router(
            _build_api(child_node, base_layout_dep=layout_dep, htmy=htmy),
            prefix=sub_url,
        )

    return api


def _build_app_tree(packages: Iterable[PackageInfo]) -> AppNode:
    """
    Returns an `AppNode` that represents the entire package tree of the application.

    Arguments:
        packages: The packages to include in the tree.
    """
    root = AppNode("/")

    for p in packages:
        root.add(p)

    return root


def _discover_app_packages(config: AppConfig) -> set[PackageInfo]:
    """
    Discovers all packages that are part of the application and returns them as a set.
    """

    @lru_cache()
    def is_excluded(path: Path) -> bool:
        """Returns whether the given file or package path should be excluded from the application."""
        return any(
            # Exclude if a path segment starts with an underscore but does not end with one.
            # Path segments that both start and end with an underscore represent path parameters!
            (p.startswith("_") and not p.endswith("_"))
            # Also exclude paths that start with a dot (virtual env, git, etc.)
            or p.startswith(".")
            for p in path.parts
        )

    return {
        PackageInfo.from_marker_file(f, config=config)
        for f in config.app_dir.rglob("*.py")
        if f.stem in module_names
        # Pass the relative path of the parent to make the best use of caching
        and not is_excluded(f.parent.relative_to(config.root_dir))
    }


def _make_api_router_for_package(pkg: PackageInfo | None, htmy: HTMY) -> APIRouter:
    """
    Creates an `APIRouter` for the given package.
    """
    api_module = None if pkg is None else pkg.import_module("api", is_api_definition)
    if api_module is None:
        return APIRouter()

    # api is either an APIRouter or a callable (is_api_definition validates that).
    api = api_module.api
    if isinstance(api, APIRouter):
        return api

    # Inspect the api callable instead of catching errors. Using a try-except with
    # a TypeError handler would hide many potential issues that would then be
    # hard to figure out without the exception trace.
    num_params = len(inspect.signature(api).parameters)
    if num_params == 0:
        api = cast(PlainAPIFactory, api)()
    elif num_params == 1:
        api = cast(RenderingAPIFactory, api)(htmy)

    if isinstance(api, APIRouter):
        return api

    raise ValueError(f"The api function of {cast(PackageInfo, pkg).package_name} must return an APIRouter.")


def _make_page_path_operation(
    *,
    layout_dep: FastAPIDependency[LayoutFactory],
    metadata_dep: FastAPIDependency[MetadataMapping | None],
    page_dep: FastAPIDependency[Any],
) -> FastAPIDependency[tuple[Component, MetadataMapping | None] | Response]:
    """Creates the path operation for a page-like route."""

    async def path_operation(
        # Start by evaluating the page dependency, it is the most likely to raise an error
        # (could even be a performance improvement strategy when returning a Response).
        page: Component | Response = Depends(page_dep),  # noqa: B008
        # Next should be the metadata dependency. It is usually relatively lightweight.
        metadata: MetadataMapping | None = Depends(metadata_dep),  # noqa: B008
        # Evaluate the layout dependency last. It's often a sequence of nested dependencies
        # and it is also unlikely to fail.
        layout: LayoutFactory = Depends(layout_dep),  # noqa: B008
    ) -> tuple[Component, MetadataMapping | None] | Response:
        if isinstance(page, Response):
            return page

        result = layout(as_component_type(page))
        # We must await here if result is an Awaitable, otherwise we would pass an
        # awaitable to htmy.page() and rendering that would fail.
        if isinstance(result, Awaitable):
            result = await result

        return result, metadata

    return path_operation



from __future__ import annotations

import inspect
import logging
from collections.abc import Callable
from dataclasses import dataclass, field
from importlib import import_module
from pathlib import Path
from typing import Any, TypeGuard, TypeVar

from .typing import ModuleName, URLSegment

_TModule = TypeVar("_TModule")


@dataclass(frozen=True, kw_only=True, slots=True)
class AppConfig:
    """
    Stores basic configuration Information about the application.
    """

    root_dir: Path
    """The root directory (as in Python import root)."""

    app_dir_name: str
    """The name of the application directory/package."""

    app_url_prefix_length: int
    """The length of the application URL prefix to remove (`len(app_dir_name) + 1`)."""

    app_dir: Path
    """Path to the application directory."""

    # Overriding __init__ and calculating inferred values automatically would serve the
    # default case well, but it would complicate things if we ever need to support
    # different application structures (for example an app in tests). So we use
    # alternative initializers instead.

    @classmethod
    def default(cls) -> AppConfig:
        """
        Creates a default instance for the standard application configuration.

        Raises:
            ValueError: If the application package cannot be determined.
        """
        root_dir = Path.cwd()
        app_dir_name = cls._find_app_package_name()
        len_app_dir_name = len(app_dir_name)

        return cls(
            root_dir=root_dir,
            app_dir_name=app_dir_name,
            # Support applications that are not wrapped in a Python package.
            # In that case app_dir_name is "".
            app_url_prefix_length=len_app_dir_name + 1 if len_app_dir_name > 0 else 0,
            app_dir=root_dir / app_dir_name,
        )

    @classmethod
    def _find_app_package_name(cls) -> str:
        """
        Attempts to find and return the application package name.

        The method uses the assumption that the application package is the one from which
        the call stack directly led to this library.

        Raises:
            ValueError: If the application package cannot be determined.
        """
        # We assume the app package is the one from which the call stack led to this library.
        lib_root = __name__.split(".")[0]
        stack = inspect.stack()
        for i in range(1, len(stack)):
            caller_frame = stack[i]
            caller_module = inspect.getmodule(caller_frame.frame)
            caller_package = caller_module.__package__ if caller_module else None
            if caller_package is None:
                break  # Couldn't find app package.

            if not caller_package.startswith(lib_root):
                # The call stack is no longer in this library, so we've found the app package.
                return caller_package

        raise ValueError("Could not determine the application package.")


@dataclass(frozen=True, kw_only=True, slots=True)
class PackageInfo:
    """
    Information about a single package.

    This class most importantly describes how a package can be imported and
    what should be the web URL of its contents.
    """

    package_dir: Path
    """
    Relative path to the package directory.

    It is basically the path matching the package name / import path.
    """

    package_name: str
    """
    The import name (path) of the package.
    """

    url: str
    """
    The web URL for the contents of the package.
    """

    def __hash__(self) -> int:
        return hash((PackageInfo, self.package_name))

    def import_module(
        self, name: ModuleName, validate: Callable[[Any], TypeGuard[_TModule]]
    ) -> _TModule | None:
        """
        Imports the module with the given name from the package.

        Arguments:
            name: The name of the module to import.
            validate: A function that validates the imported module.

        Returns:
            The imported module or `None` if the module does not exist.

        Raises:
            ValueError: If the module is invalid.
        """
        # Check if the module exists. Trying to import it and catching the exception would hide possible
        # import error that occur in within the module, which is undesired because it would be hard for
        # users to find why an API is not registered.
        if not (self.package_dir / f"{name}.py").is_file():
            return None

        # Support applications that are not wrapped in a Python package.
        # In that case `self.package_name` is ".".
        import_name = name if self.package_name == "." else f"{self.package_name}.{name}"
        try:
            module = import_module(import_name)
        except Exception:
            import traceback

            # Handle potential misconfigurations with a simple warning, instead of an exception.
            logger = logging.getLogger("holm")
            logger.warning(f"Failed to import module {name} at: {import_name}")
            logger.warning(traceback.format_exc())
            return None

        if not validate(module):
            raise ValueError(f"Invalid module: {name}")

        return module

    @classmethod
    def from_marker_file(cls, file_path: Path, *, config: AppConfig) -> PackageInfo:
        """
        Creates a `PackageInfo` instance from a so called "marker" that is located in the package.
        """
        package_dir = file_path.relative_to(config.root_dir).parent
        package_dir_str = str(package_dir)
        package_name = package_dir_str.replace("/", ".")
        # Create the base version of the URL
        url = "/" if package_dir_str == "." else f"/{package_dir_str[config.app_url_prefix_length :]}"
        # Technically there can be packages/directories with names like `{name}`. They can be
        # directly used as URL segments and they could even be imported with `import_module()`,
        # but they are invalid package names and static code analysis tools rightly complain
        # about them, so we need an alternative option.
        # That option is to use the `_name_` pattern instead and convert such package/directory
        # names to the `{name}` format here and use that as the URL segment.
        url = "/".join(
            (f"{{{p[1:-1]}}}" if len(p) > 2 and p[0] == p[-1] == "_" else p for p in url.split("/"))
        )
        return cls(package_dir=package_dir, package_name=package_name, url=url)

    def __lt__(self, other: PackageInfo) -> bool:
        return self.package_name < other.package_name


@dataclass(frozen=True, slots=True)
class AppNode:
    """
    The description of an application tree.
    """

    url: str
    """The full URL of this node."""

    subtree: dict[URLSegment, AppNode] = field(default_factory=dict)
    """
    Maps sub-URLs to child nodes.
    """

    _package: PackageInfo | None = None
    """The package that belongs to this node."""

    @property
    def last_url_segment(self) -> URLSegment:
        """The last URL segment of this node."""
        return f"/{self.url.rsplit('/', 1)[1]}"

    @property
    def package(self) -> PackageInfo | None:
        """The package that belongs to this node."""
        return self._package

    def add(self, item: PackageInfo) -> None:
        """
        Adds the given package to this node.

        If the package belongs in the subtree and not directly to this node, then the
        entire subtree is created automatically.
        """
        self_url = self.url
        if not item.url.startswith(self_url):
            raise ValueError(f"{item.url} must start with {self_url}")

        if item.url == self_url:
            if self._package is not None:
                raise ValueError(
                    (
                        f"Package already set for {self_url}: "
                        f"current is {self._package.package_name} "
                        f"new is {item.package_name}"
                    )
                )

            # Class is frozen so this is only way to set _package.
            object.__setattr__(self, "_package", item)
            return

        if self_url == "/":  # Root node
            sub_url = item.url
            url_prefix = ""
        else:
            sub_url = item.url.removeprefix(self_url)
            url_prefix = self_url

        first_segment: URLSegment = f"/{sub_url.split('/', 2)[1]}"  # First segment after leading slash
        child = self.subtree.get(first_segment)
        if child is None:
            # The URL for this node must be constructed this way to not skip intermediate nodes
            # that may not have a discovered package.
            child = AppNode(f"{url_prefix}{first_segment}")
            self.subtree[first_segment] = child

        child.add(item)



from collections.abc import Callable, Coroutine
from typing import Any, TypeAlias, TypeVar

from fastapi import Request

_T = TypeVar("_T")

SyncFastAPIDependency: TypeAlias = Callable[..., _T]
AsyncFastAPIDependency: TypeAlias = Callable[..., Coroutine[None, None, _T]]

FastAPIDependency: TypeAlias = SyncFastAPIDependency[_T] | AsyncFastAPIDependency[_T]
"""FastAPI dependency typing."""

FastAPIErrorHandler: TypeAlias = Callable[[Request, Exception], Coroutine[None, None, Any]]
"""Generic FastAPI error handler typing."""



from typing import Any, Protocol, TypeAlias

from holm.fastapi import FastAPIDependency

SubmitHandler: TypeAlias = FastAPIDependency[Any]
"""FastAPI dependency that returns the properties for the layout that wraps it
(for example a `Component`).

If the submit handler is not wrapped by a layout, then it must return a `Component`
"""


class SubmitHandlerOwner(Protocol):
    """Protocol definition for objects (usually modules) that define a submit handler."""

    @property
    def handle_submit(self) -> SubmitHandler: ...


def get_submit_handler(obj: Any) -> SubmitHandler | None:
    """Loads and returns the submit handler if it exists in the given object."""
    submit_handler = getattr(obj, "handle_submit", None)
    return submit_handler if callable(submit_handler) else None
